@@ -334,6 +334,47 @@ public class JavaBridge {
 	  };
 	}
 
+	private boolean loadJni() {
+		try {
+			String javaHome = System.getProperty ("java.home");
+			String jvmArch = System.getProperty ("sun.arch.data.model");
+			String osArch = System.getProperty ("os.arch");
+			String fileSep = System.getProperty ("file.separator");
+			String libPath = javaHome + "/lib/ext/libjava-access-bridge-jni.so";
+
+			if ("64".equals(jvmArch)) {
+				String libPath64 = javaHome + "/lib/ext/" + osArch + "/libjava-access-bridge-jni.so";
+				java.io.File fileLib64 = new java.io.File( libPath64.replace('/', fileSep.charAt(0)) );
+				if (fileLib64.exists()) {
+					libPath = libPath64;
+				}
+			}
+			
+			if (!libPath.equals("")) {
+				libPath = libPath.replace('/', fileSep.charAt(0));
+				System.load (libPath);
+				
+				if (debugFlag) {
+					System.err.println ( libPath + " loaded" );
+				}
+			} else {
+				if (debugFlag) {
+					System.err.println ( "Error: jni library not found" );
+				}
+
+				return false;
+			}
+		}
+		catch (Throwable ex)
+		{
+			System.err.println(ex+" caught.");
+			ex.printStackTrace();
+			return false;
+		}
+
+		return true;
+	}
+
 	public static void main(String args[]) {
 
 		new JavaBridge();
@@ -347,18 +388,6 @@ public class JavaBridge {
 		if (debugFlag)
 			System.err.println ("Java Accessibility Bridge for GNOME loaded.\n");
 
-        try {
-            String javaHome = System.getProperty ("java.home");
-            System.load ( javaHome + "/lib/ext/libjava-access-bridge-jni.so" );
-			if (debugFlag)
-				System.err.println ( javaHome + "/lib/ext/libjava-access-bridge-jni.so loaded" );
-        }
-        catch (Exception ex)
-        {
-            System.err.println(ex+" caught.");
-            ex.printStackTrace();
-        }
-
 		// Not sure what kind of arguments should be sent to ORB
 		String vm_rev = System.getProperty("java.version");
 
@@ -1111,6 +1140,10 @@ public class JavaBridge {
 				return false;
 			}
 
+			if (! loadJni() ) {
+				return false;
+			}
+
 			if (AccessUtil.getRegistryObject() == null)
 				return false;
 
