@@ -198,7 +198,8 @@ public class TextImpl extends UnknownImpl implements TextOperations {
     }
 
 	private StringStruct getLine (javax.accessibility.AccessibleText accText,
-				      int offset) {
+				      int offset,
+				      TEXT_BOUNDARY_TYPE type) {
 	    //System.err.println("getLine ( " + offset + " )");
 		String string = "";
 		StringStruct lineStruct = new StringStruct ();
@@ -213,11 +214,16 @@ public class TextImpl extends UnknownImpl implements TextOperations {
                 bounds = accText.getCharacterBounds (offset - 1);
             }
 
-            if (start > 0) do {
-                currentBounds = accText.getCharacterBounds (--start);
-            } while (start > 0 && 
-                     overlapsVertically( bounds, currentBounds ));
-
+            while (start > 0) {
+		currentBounds = accText.getCharacterBounds (--start);
+		if (overlapsVertically( bounds, currentBounds )) {
+		    continue;
+		}
+		else {
+		    ++start;
+		    break;
+		}
+	    }
             currentBounds = bounds;
             while (end < lastOffset && 
                    overlapsVertically( bounds, currentBounds )) {
@@ -229,6 +235,10 @@ public class TextImpl extends UnknownImpl implements TextOperations {
         catch ( Exception e ) {
             e.printStackTrace();
         }
+	if (type == TEXT_BOUNDARY_TYPE.TEXT_BOUNDARY_LINE_END) {
+	    if (start > 0)
+		start--;
+	}
         lineStruct.startOffset = start;
         lineStruct.endOffset = end;
         if (string == "") 
@@ -375,7 +385,7 @@ public class TextImpl extends UnknownImpl implements TextOperations {
                 boundary_type = boundaryTypeConvert (type); 
                 text = accText.getAtIndex(boundary_type, offset);
                 if ( text == null )
-                    return getLine (accText, offset);
+                    return getLine (accText, offset, type);
             } else {	 
                 boundary_type = boundaryTypeConvert (type); 
                 text = accText.getAtIndex(boundary_type, offset);
@@ -538,6 +548,94 @@ public class TextImpl extends UnknownImpl implements TextOperations {
 
 	}
 
+        private Object toJavaTextAttribute (String name)
+	{
+            if ( name.equals( "family-name" ) ) {
+		return "family";
+	    }
+            else if ( name.equals( "style" ) ) {
+                return "italic";
+            }
+	    else if ( name.equals( "weight" ) ) {
+		return "bold";
+	    }
+	    else if ( name.equals ( "fg-color" ) ) {
+		return "foreground";
+	    }
+	    else if ( name.equals ( "bg-color" ) ) {
+		return "background";
+	    }
+	    else return name;
+	}
+
+        private String fromJavaTextAttribute (Object name, Object value)
+	{
+	    String result = "";
+            String nameStr = name.toString();
+            String valueStr = value.toString();
+            // System.out.println("name = " + nameStr + "; value = " + valueStr );
+            if ( nameStr.equals( "family" ) ) {
+                result = "family-name:" + valueStr;
+            }
+            else if ( nameStr.equals( "size" ) ) {
+                result = nameStr + ":" + valueStr;
+            }
+            else if ( nameStr.equals( "italic" ) ) {
+                if ( Boolean.valueOf( valueStr ).booleanValue() ) {
+                    result = "style:italic";
+                }
+                else {
+                    result = "style:normal";
+                }
+            }
+            else if ( nameStr.equals( "bold" ) ) {
+                if ( Boolean.valueOf( valueStr ).booleanValue() ) {
+                    result = "weight:700";
+                }
+                else {
+                    result = "weight:400";
+                }
+            }
+            else if ( nameStr.equals( "foreground" ) || nameStr.equals( "color" ) ) {
+                if ( value instanceof java.awt.Color ) {
+                    Color c = (Color)value;
+                    result = "fg-color:";
+                    result += (c.getRed() * 256) + c.getRed();
+                    result += ",";
+                    result += (c.getGreen() * 256) + c.getGreen();
+                    result += ",";
+                    result += (c.getBlue() * 256) + c.getBlue();
+                }
+            }
+            else if ( nameStr.equals( "background" ) ) {
+                if ( value instanceof java.awt.Color ) {
+                    Color c = (Color)value;
+                    result = "bg-color:";
+                    result += (c.getRed() * 256) + c.getRed();
+                    result += ",";
+                    result += (c.getGreen() * 256) + c.getGreen();
+                    result += ",";
+                    result += (c.getBlue() * 256) + c.getBlue();
+                }
+            }
+            else if ( nameStr.equals( "underline" ) ) {
+                result = "underline:";
+                if ( Boolean.valueOf( valueStr ).booleanValue() ) {
+                    result += "single";
+                }
+                else {
+                    result += "none";
+                }
+            }
+            else if ( nameStr.equals( "strikethrough" ) ) {
+                result = "strikethrough:" + Boolean.valueOf( valueStr ).booleanValue();
+            }
+            else {  // Default to Strings
+                result = nameStr + ":" + valueStr;
+            }
+	    return result;
+	}
+
 	private String attributeSetConvert (javax.swing.text.AttributeSet attributes) {
 		// TODO: harmonize attribute names with atk-bridge/at-spi
 		String result = "";
@@ -549,69 +647,7 @@ public class TextImpl extends UnknownImpl implements TextOperations {
             //                     name.getClass().getName() );
             // System.out.println("attributeSetConvert: value is type : " +
             //                     value.getClass().getName() );
-
-            String nameStr = name.toString();
-            String valueStr = value.toString();
-            // System.out.println("name = " + nameStr + "; value = " + valueStr );
-            if ( nameStr.equals( "family" ) ) {
-                result += "family-name:" + valueStr;
-            }
-            else if ( nameStr.equals( "size" ) ) {
-                result += nameStr + ":" + valueStr;
-            }
-            else if ( nameStr.equals( "italic" ) ) {
-                if ( Boolean.valueOf( valueStr ).booleanValue() ) {
-                    result += "style:italic";
-                }
-                else {
-                    result += "style:normal";
-                }
-            }
-            else if ( nameStr.equals( "bold" ) ) {
-                if ( Boolean.valueOf( valueStr ).booleanValue() ) {
-                    result += "weight:700";
-                }
-                else {
-                    result += "weight:400";
-                }
-            }
-            else if ( nameStr.equals( "foreground" ) || nameStr.equals( "color" ) ) {
-                if ( value instanceof java.awt.Color ) {
-                    Color c = (Color)value;
-                    result += "fg-color:";
-                    result += (c.getRed() * 256) + c.getRed();
-                    result += ",";
-                    result += (c.getGreen() * 256) + c.getGreen();
-                    result += ",";
-                    result += (c.getBlue() * 256) + c.getBlue();
-                }
-            }
-            else if ( nameStr.equals( "background" ) ) {
-                if ( value instanceof java.awt.Color ) {
-                    Color c = (Color)value;
-                    result += "bg-color:";
-                    result += (c.getRed() * 256) + c.getRed();
-                    result += ",";
-                    result += (c.getGreen() * 256) + c.getGreen();
-                    result += ",";
-                    result += (c.getBlue() * 256) + c.getBlue();
-                }
-            }
-            else if ( nameStr.equals( "underline" ) ) {
-                result += "underline:";
-                if ( Boolean.valueOf( valueStr ).booleanValue() ) {
-                    result += "single";
-                }
-                else {
-                    result += "none";
-                }
-            }
-            else if ( nameStr.equals( "strikethrough" ) ) {
-                result += "strikethrough:" + Boolean.valueOf( valueStr ).booleanValue();
-            }
-            else {  // Default to Strings
-                result += nameStr + ":" + valueStr;
-            }
+			result += fromJavaTextAttribute (name, value);
 			if (names.hasMoreElements ()) result += "; ";
 		}
         // System.out.println("attributeSetConvert : returns : >>" + result + "<<" );
@@ -829,12 +865,54 @@ public class TextImpl extends UnknownImpl implements TextOperations {
 		return false;
 	}
 
-    public void getRangeExtents (int startOffset, int endOffset, 
-                                 org.omg.CORBA.IntHolder x, org.omg.CORBA.IntHolder y, 
-                                 org.omg.CORBA.IntHolder width, 
-                                 org.omg.CORBA.IntHolder height, short coordType) {
+   public void getRangeExtents (int startOffset, int endOffset,
+                                 org.omg.CORBA.IntHolder x,
+                                 org.omg.CORBA.IntHolder y,
+                                 org.omg.CORBA.IntHolder width,
+                                 org.omg.CORBA.IntHolder height,
+                                 short coordType) {
+        /* coordType:
+           If 0, the above coordinates are reported in pixels relative
+           to corner of the screen; if 1, the coordinates are reported
+           relative to the corner of the containing toplevel window.
+        */
+        if (coordType != 0 && coordType != 1) {
+            return;
+        }
+
+        if (accText instanceof AccessibleExtendedText) {
+            // getTextBounds returns bounds relative to the parent container.
+            AccessibleExtendedText accExtendedText = (AccessibleExtendedText)accText;
+            Rectangle rect = accExtendedText.getTextBounds(startOffset, endOffset);
+            if (rect == null) {
+                return;
+            }
+            AccessibleContext accParent =
+                ac.getAccessibleParent().getAccessibleContext();
+            AccessibleComponent accComponent = (AccessibleComponent)accParent;
+            Point parentLocation = accComponent.getLocationOnScreen();
+
+            if (coordType == 0) {
+                // getTextBounds returns a location relative to the parent
+                // container. Add the location of the parent container relative
+                // to the top-right corner of the screen.
+                x.value = rect.x + parentLocation.x;
+                y.value = rect.y + parentLocation.y;
+                width.value = rect.width;
+                height.value = rect.height;
+            } else {
+                // Same as for (coordType == 0) above, but subtract the location
+                // of the top-level container.
+                Point topLocation = getToplevelLocationOnScreen();
+                x.value = rect.x + parentLocation.x - topLocation.x;
+                y.value = rect.y + parentLocation.y - topLocation.y;
+                width.value = rect.width;
+                height.value = rect.height;
+            }
+        }
     }
 
+
     public org.GNOME.Accessibility.TextPackage.Range[] getBoundedRanges (
                                 int x, int y, int width, int height, short coordType, 
                                 org.GNOME.Accessibility.TEXT_CLIP_TYPE xClipType, 
@@ -846,11 +924,84 @@ public class TextImpl extends UnknownImpl implements TextOperations {
 
 		return "";
     }
+        public String[] getDefaultAttributeSet ()  {
+
+                return null;
+    }
+
+        public String[] getAttributeRun(int p1,org.omg.CORBA.IntHolder p2,org.omg.CORBA.IntHolder p3,boolean p4) {
+                return null;
+    }
+
+        public String getAttributeValue (int offset,
+					 java.lang.String name,
+					 org.omg.CORBA.IntHolder startOffset,
+					 org.omg.CORBA.IntHolder endOffset,
+					 org.omg.CORBA.BooleanHolder defined) {
+        int start = offset;
+        int end = offset;
+        int last = -1;
+	Object jname = toJavaTextAttribute (name);
+        try {
+            if (accText != null) {
+                javax.swing.text.AttributeSet attributes = 
+                    accText.getCharacterAttribute(offset);
+		if (!attributes.isDefined (jname)) {
+		    defined.value = false;
+		    return "";
+		} 
+		else {
+		    defined.value = true;
+		}
+                last = accText.getCharCount ();
+
+                javax.swing.text.AttributeSet tmp_attrs;
+                start--;
+                if ( start > 0 ) {
+                    tmp_attrs = accText.getCharacterAttribute(start);
+                    while ( start > 0 &&
+                            (( attributes == null && tmp_attrs == null) ||
+                             ( attributes != null && tmp_attrs != null &&
+                               attributes.getAttribute (name) == tmp_attrs.getAttribute (jname) ) ) ) {
+                        tmp_attrs = accText.getCharacterAttribute(--start);
+                    }
+                }
+
+                end++;
+                if ( end < last ) {
+                    tmp_attrs = accText.getCharacterAttribute(end);
+                    while ( end < last &&
+                            (( attributes == null && attributes == tmp_attrs ) ||
+                             ( attributes != null && tmp_attrs != null &&
+                               attributes.getAttribute (name) == tmp_attrs.getAttribute (jname) ) ) ) {
+                        tmp_attrs = accText.getCharacterAttribute(++end);
+                    }
+                }
+                startOffset.value = ++start;
+                endOffset.value = end;
+                if ( attributes != null ) {
+                    return fromJavaTextAttribute (jname, attributes.getAttribute (jname));
+		}
+                else
+                    return "";
+            }
+        }
+        catch ( Exception e ) {
+	    defined.value = false;
+            e.printStackTrace();
+        }
+        startOffset.value = offset;
+        if ( last < 0 )
+            endOffset.value = offset + 1;
+        else 
+            endOffset.value = last;
+        return "";
+	}
+
 	public void unImplemented () {}
 	public void unImplemented2 () {}
 	public void unImplemented3 () {}
 	public void unImplemented4 () {}
-	public void unImplemented7 () {}
 
     private String typeToString( org.GNOME.Accessibility.TEXT_BOUNDARY_TYPE type ) {
         String rval;
